#!/usr/bin/env python

#-------------------------------------------------------------------
# Objectives
#-------------------------------------------------------------------
# The script will ask for a revision number. This revision number 
# can be provided through the terminal or it can be given through 
# an argument at the time the script is ran:
# > python Make_orchidee_pft_defs.py -rev XXXX
# or
# > python Make_orchidee_pft_defs.py -r XXXX
#
# Use parameter tables/database to produce up-to-date orchidee_pft.defs 
# for different simulation set-ups.
# 
# The script will write the orchidee_pft.def files directly into the
# PARAM folder of the different default experiments (ENSEMBLE, FG1, 
# FG2, FG3, SPINUP, ...)
#
# A variable can be added, removed and then added again. When this
# is the case, parameter values for model versions prior to its second
# addition will no longer be considered. Parameter values will have
# to be re-enterd. The reasoning behind this decision is that there
# probably was a good reason to remove the parameter. If its needed
# to add it again, other values than those before its removal will
# likley be used. If this not the case, it should probably not have 
# been removed in the first place. Or it was used a work around to
# show the values of a limited set of PFTs.
#
# The script assumes that the variables in the "removed" and "added"
# list are sorted in ascending order accorind to their revision number.
# The code has not been tested for unsorted lists.
#
# CODING GUIDELINE: variable names written in capitals are variables
# that will end up in the run.defs

#-------------------------------------------------------------------
# Load modules and functions definitions
#-------------------------------------------------------------------
import numpy as np
import sys
 
#-------------------------------------------------------------------
# Some statements in this script are specific for Python2.
# Confirm we are running a good version of Python.
#-------------------------------------------------------------------

if sys.version_info > (3, 0):
    print('Python version is probably too high.  Try 2.7')
    print(sys.version)
    sys.exit()

#-------------------------------------------------------------------
# List the experiments that will make use of the orchidee_pft
# files generated by this script. The ENSEMBLE experiment has its
# PARAM folder in SPINUP. No need to add the orchidee_pft.defs 
# in there. 	
#-------------------------------------------------------------------
experiments = ['OOL_SEC',\
               'OOL_SEC_STO_FG1trans',\
               'OOL_SEC_STO_FG1transnd',\
               'OOL_SEC_STO_FG2',\
               'OOL_SEC_STO_FG2nd',\
               'OOL_SEC_STO_FG3',\
               'OOL_SEC_STO_FG3nd',\
               'OOL_SEC_STO_FG4',\
               'OOL_SEC_STO_FG5',\
               'OOL_SEC_STO_FG8nd',\
               'SPINUP',\
               'SPINUP_ANALYTIC_FG1',\
               'SPINUP_ANALYTIC_FG1nd',\
               'ENSEMBLE']
n_experiments = np.shape(experiments)[0]

#-------------------------------------------------------------------
# Import tables with parameter values
#-------------------------------------------------------------------
# Convert into more workable variables. parameter,
# removed and added are tables that are defined in 
# processed_parameters.py.
from processed_parameters import *
n_records = np.shape(parameter)[0]
n_removed = np.shape(removed)[0]
n_added = np.shape(added)[0]
date = []         
mtc = []
pft = []
species = []
parname = []
for irecord in range(n_records):
    # Copy integers
    date.append(parameter[irecord][0])
    # Copy strings into lists
    mtc.append(parameter[irecord][1]) 
    pft.append(parameter[irecord][2])
    species.append(parameter[irecord][3])
    parname.append(parameter[irecord][4])
#endfor

# Extract a list with unique parameter names
parameter_names = np.unique(parname)
n_parameters = len(parameter_names)
parameter_units = []

# Find the matching units for the parameters
for iparameter in range(n_parameters):
    found = 0
    irecord = 0
    while found == 0:
        if parname[irecord] == parameter_names[iparameter]:
            parameter_units.append(parameter[irecord][5])
            found = 1        
        else:
            irecord = irecord+1
        #endif
    #endwhile
#endfor

#-------------------------------------------------------------------
# Make run.defs for different configurations
#-------------------------------------------------------------------
# Get the ORCHIDEE revision from argument or by reading from terminal
if len(sys.argv) > 2:
  if sys.argv[1]=='-rev' or sys.argv[1]=='-r':
    orchidee_revision=sys.argv[2]
  else:
    print('Arguments used are',sys.argv)
    print('Error in syntax for argument. Use -r or -rev followed by a revision number.')
    sys.exit()
  #endif
else:
  print('Enter ORCHIDEE revision number (4 digits):')
  orchidee_revision = input()
#endif

print('Make orchidee_pft.def file for following ORCHIDEE revision:',orchidee_revision)

# Now that we know the revision number we can create a variable that
# contains the range in revisions numbers we have to consider when
# making the run.def
parameter_minrev = np.zeros(n_parameters)
parameter_maxrev = np.zeros(n_parameters)+orchidee_revision

# Use the revision numbers in the added and removed lists
# to built the range of revision numbers that should be
# considered when making the run.def
for iadded in range(n_added):
    added_to_revision = int(added[iadded][1].replace("r",""))
    index = np.where(parameter_names==added[iadded][0])
    if added_to_revision < parameter_maxrev[index] and \
       added_to_revision > parameter_minrev[index]:
        # The parameter was added after the start of using
        # this script and before the revision number we want
        # to a make a run.def for. The parameter should be
        # added. If this parameter is added more than once
        # the revision number send to minrev might be updated.
        # If this parameter was added a second time with a
        # larger revision number nothing needs to be done.
        parameter_minrev[index] = added_to_revision
    #endif
#endfor
for iremoved in range(n_removed):
    removed_from_revision = int(removed[iremoved][1].replace("r",""))    
    index = np.where(parameter_names==removed[iremoved][0])
    if removed_from_revision < parameter_maxrev[index] and \
       removed_from_revision > parameter_minrev[index]:
        # The parameter was removed in the range of revision
        # numbers we have to consider according to the
        # information from the "added" list.
        parameter_names[index] = 'EMPTY'
        parameter_minrev[index] = 0
        parameter_maxrev[index] = 0
    #endif
#endfor

# Quick error checking
for iparameter in range(n_parameters):
    if parameter_maxrev[iparameter] > 0:
        print 'Variable: ',parameter_names[iparameter], \
            'Lowest/highest revision number: ', int(parameter_minrev[iparameter]), \
            '/', int(parameter_maxrev[iparameter])
        if parameter_minrev[iparameter] > parameter_maxrev[iparameter]:
            print ('Error for parameter: ', parameter_names[index])
            print ('Lowest exceeds highest revision number')
        #endif
    #endif
#endfor



# Repeat the process for different PFT configurations.
# Each configuration has been defined in a separate file.
for i in range(0,6):
        
    #------------------------------------------
    # Get the configuration of the PFTs
    #------------------------------------------
    outlist2 = []
    # Initialize the run.def
    if (i==0):
        from config3PFTs import config, NAGEC
        npft_expected = 3
    elif (i==1):
        from config13PFTs import config, NAGEC
        npft_expected = 13    
    elif (i==2):
        from config15PFTs import config, NAGEC
        npft_expected = 15
    elif (i==3):
        from config28PFTs import config, NAGEC
        npft_expected = 28
    #elif (i==4):
        #from config37PFTs import config, NAGEC
        #npft_expected = 37
    elif (i==4):
        from config39PFTs import config, NAGEC
        npft_expected = 39
    elif (i==5):
        from config64PFTs import config, NAGEC
        npft_expected = 64
    #endif
    print('Working on the run.def with '+str(npft_expected)+' PFTs')
            
    #------------------------------------------
    # Write the model settings
    #------------------------------------------
    # Initialize administrative variables. These
    # variables are hard coded and therefore always
    # written. They do not depend on the list of
    # removed variables.
    n_vegetation_classes = np.shape(config)[0]
    has_agec = np.zeros(shape=(n_vegetation_classes),dtype=np.str)
        
    # Copy values in more workable variables
    for iclass in range(n_vegetation_classes):        
        # Copy the integers into new variables
        has_agec[iclass] = config[iclass][5]
    #endfor

    #------------------------------------------    
    # NVM and NVMAP
    #------------------------------------------
    text='\n'+'#'+75*'*'+'\n'+\
        '#'+        '            Set-up age classes'+'\n'+\
        '#'+75*'*'+'\n'*2+\
        '# number of PFTs required in the PFT map and number of PFTs in the run.def'
    outlist2.append(text)
    npft=0
    npft_noagec=0
    for iclass in range(n_vegetation_classes):
        if has_agec[iclass].strip()=='y':
            npft=npft+int(NAGEC)
            npft_noagec=npft_noagec+1
        else:
            npft=npft+1
            npft_noagec=npft_noagec+1
        #endif
    #endfor

    text='NAGEC='+str(NAGEC)
    outlist2.append(text) 
    text='NVMAP='+str(npft_noagec)
    outlist2.append(text) 
    text='NVM = '+ str(npft)
    outlist2.append(text)
    outlist2.append('\n')

    #------------------------------------------    
    # PFT_TO_MTC
    #------------------------------------------
    text='# Link PFT''s to MTC''s'    
    outlist2.append(text)
    
    # Read from configuration file
    PFT_TO_MTC = np.zeros(shape=(n_vegetation_classes),dtype=np.int)
    for iclass in range(n_vegetation_classes):        
        # Copy the integers into a new variable
        PFT_TO_MTC[iclass] = config[iclass][4]
    #endfor

    npft=0    
    for iclass in range(n_vegetation_classes):
        mtc_value=PFT_TO_MTC[iclass]
        if has_agec[iclass].strip()=='y':
            loopN=int(NAGEC)
        else:
            loopN=1
        #endif    
                
        for s in range(loopN):
            npft=npft+1
            text='PFT_TO_MTC__'+str(npft).zfill(2)+'='+ str(mtc_value)
            outlist2.append(text)
        #endfor
    #endfor
    outlist2.append('\n')

    #------------------------------------------    
    # AGEC_GROUP
    #------------------------------------------
    text='# Link PFT to an age class group'    
    outlist2.append(text)
    npft=0
    npft_noagec=0
    for iclass in range(n_vegetation_classes):
        if has_agec[iclass].strip()=='y':
            loopiclass=int(NAGEC)
        else:
            loopiclass=1
        #endif
        
        for s in range(loopiclass):
            npft=npft+1        
            if s == 0:
                npft_noagec=npft_noagec+1
                text='AGEC_GROUP__'+str(npft).zfill(2)+'='+str(npft_noagec)
            else:
                npft_noagec=npft_noagec
                text='AGEC_GROUP__'+str(npft).zfill(2)+'='+str(npft_noagec)
            #endif
            outlist2.append(text)
        #endfor
    #endfor
    outlist2.append('\n')

    #------------------------------------------    
    # PFT_NAME
    #------------------------------------------
    text='# Description of PFTs'
    outlist2.append(text)
    npft=0 
    for iclass in range(n_vegetation_classes):
        species_name=config[iclass][2]+config[iclass][3]
        if has_agec[iclass].strip()=='y':
            loopiclass=int(NAGEC)
            for s in range(loopiclass):
                npft=npft+1
                text='PFT_NAME__' +str(npft).zfill(2)+'='+species_name+'_'\
                    'age'+str(s+1).zfill(2)
                outlist2.append(text)
            #endfor
        else:
            npft=npft+1
            text='PFT_NAME__' +str(npft).zfill(2)+'='+species_name
            outlist2.append(text)
        #endif
    #endfor
    outlist2.append('\n')

    #------------------------------------------    
    # PARAMETERS FROM PROCESSED TABLE
    #------------------------------------------
    # Extract parameter by parameter. The parameters
    # treated in the code below depend on the "added"
    # and "removed" lists.
    for iparameter in range(n_parameters):
        # This parameter was removed from the parameter list for
        # this ORCHIDEE revision.
        if (parameter_names[iparameter]=='EMPTY'): continue
        header = 0
        npft=0
        
        # Extract PFT by PFT
        for iclass in range(n_vegetation_classes):
            value = 'EMPTY'
            stored_rev = 0
            
            # Find the parameter value
            n_records = np.shape(parameter)[0]
            for irecord in range(n_records):
                parrev = int(parameter[irecord][7].replace("r",""))
                if parname[irecord]==parameter_names[iparameter] and \
                   mtc[irecord]==config[iclass][1] and \
                   pft[irecord]==config[iclass][2] and \
                   species[irecord]==config[iclass][3] and \
                   parrev >= parameter_minrev[iparameter] and \
                   parrev <= parameter_maxrev[iparameter]:
                    if stored_rev == 0:
                        # Take the first value in the parameter file
                        # Remove leading r and convert to integer
                        value = parameter[irecord][6]
                        stored_rev = parrev
                        stored_date = date[irecord]
                    else:
                        # Should the first value be overwritten?
                        # It is possible that we have different values
                        # for the same revision but of different dates
                        # take the values of the most recent date.
                        # Remove leading r and convert to integer
                        if parrev <= parameter_maxrev[iparameter] and \
                           parrev >= stored_rev and \
                           date[irecord] > stored_date:
                            # Replace the parameter values by a more 
                            # recent estimate
                            value = parameter[irecord][6]
                            stored_rev = parrev
                            stored_date = date[irecord]
                        #endif
                    #endif
                #endif
            #endfor
          
            # If at least one valid value was found write this variable to the run.def
            if stored_rev > 0 and header == 0:  
                text='# '+parameter_names[iparameter]+'('+parameter_units[iparameter] +')'
                outlist2.append(text)
                header = 1
            #endif

            # Write to the run.def
            if has_agec[iclass].strip()=='y':
                loopiclass=int(NAGEC)
            else:
                loopiclass=1
            #endif
                        
            for s in range(loopiclass):
                npft=npft+1
                if value != 'EMPTY':
                    text=parameter_names[iparameter]+'__'+str(npft).zfill(2)+'='\
                          +str(value)
                    outlist2.append(text)
                #endif
            #endfor
        #endfor
        outlist2.append('\n') 
    #endfor

    #------------------------------------------    
    # PFT fractions
    #------------------------------------------
    text='\n'+'#'+75*'*'+'\n'+\
        '#'+        '            PFT fractions'+'\n'+\
        '#'+75*'*'+'\n'*2+\
        '# The fraction of vegetation is read from the restart file or a map'+'\n'+\
        '# If it is not found in one of those sources we will use the values'+'\n'+\
        '# provided here'+'\n'
    outlist2.append(text)
    suboutlist = []
    npft=0
    for iclass in range(n_vegetation_classes):
        if has_agec[iclass].strip()=='y':
            loopiclass=int(NAGEC)
        else:
            loopiclass=1
        #endif
        
        for s in range(loopiclass):
            npft=npft+1
            if (s == 0):
                if (npft_expected == 3):
                    # The test-case with only 2 identical PFTs and bare soil should
                    # have different VEGMAX for the otherwise identical PFTs. Else
                    # the test case won't test the effect of veget_max.
                    if (npft==1):
                        text1='SECHIBA_VEGMAX__'+str(npft).zfill(2)+'='+str(0.0)
                    if (npft==2):
                        text1='SECHIBA_VEGMAX__'+str(npft).zfill(2)+'='+str(0.2)
                    if (npft==3):
                        text1='SECHIBA_VEGMAX__'+str(npft).zfill(2)+'='+str(0.8)
                    #endif
                else:    
                    # Calculate an identical fraction for each PFT (in case age classes
                    # are used, only the youngest age class of each species group will
                    # get a VEGMAX
                    text1='SECHIBA_VEGMAX__'+str(npft).zfill(2)+'='+str(1.0/npft_noagec)
                #endif
            else:
                text1='SECHIBA_VEGMAX__'+str(npft).zfill(2)+'='+str(0.0)
            #endif
            outlist2.append(text1)
        #endfor
    #endfor
    outlist2.append('\n')
    outlist2.extend(suboutlist)        

    #------------------------------------------    
    # Write run.def to all the experiments
    #------------------------------------------
    # Set name of the run.def
    if (npft==npft_expected):
        for i_exp in range(n_experiments): 
            outputfile = './../'+experiments[i_exp]+'/PARAM/orchidee_pft.def_'+str(npft)+'pft.'+\
                         str(NAGEC)+'ac'
            fob = open(outputfile,'w')
            for r in outlist2:
                fob.write(r+'\n')
            #endfor
            fob.close()
        #endfor
    else:
        print('The script proposes a run.def with '+str(npft)+' PFTs')
        print('You expected '+str(npft_expected)+' PFTs')
        print('ERROR: check the configuration file')  
    #endif

#endfor - ORCHIDEE_OL configuration
